// Author: Josiah Manson

#include "pointset.h"
#include "parse.h"
#include <float.h>
#include <cstring>
#include <limits>
#include <cstdlib>

static void findExtents(vector<OrientedPoint> &pts, vect3f &mine, vect3f &maxe)
{
	if (pts.empty())
		return;

	mine = maxe = pts[0].pos;

	for (int k = 1; k < pts.size(); k++)
	{
		for (int i = 0; i < 3; i++)
		{
			mine[i] = min(mine[i], pts[k].pos[i]);
			maxe[i] = max(maxe[i], pts[k].pos[i]);
		}
	}
}

void PointSet:: loadObj(string filename)
{
	// first load the data
	points.clear();
	vector<vect3f> vertices;

	char line_raw[256];
	ifstream file(filename.c_str());
	if (file.fail())
		return;

	int tri_num = 0;
	int vert_num = 0;
	for (;;)
	{
		file.getline(line_raw, 256);
		if (file.eof())
			break;

		string line(line_raw);
		vector<string> tokens;
		parseString(&tokens, line, " \t");
		removeEmptyStrings(&tokens);

		if (tokens.size() < 4)
			continue;

		if (tokens[0] == "v")
		{
			vert_num++;
			if (vert_num % 10000 == 0)
				printf("read vertex %d\n", vert_num);

			// do it
			vect3f v;

			v[0] = atof(tokens[1].c_str());
			v[1] = atof(tokens[2].c_str());
			v[2] = atof(tokens[3].c_str());

			vertices.push_back(v);
		}
		else if (tokens[0] == "f")
		{
			tri_num++;
			if (tri_num % 10000 == 0)
				printf("read triangle %d\n", tri_num);

			// do it
			const int i0 = atoi(tokens[1].c_str()) - 1;
			const int i1 = atoi(tokens[2].c_str()) - 1;
			const int i2 = atoi(tokens[3].c_str()) - 1;

			const vect3f &v0 = vertices[i0];
			const vect3f &v1 = vertices[i1];
			const vect3f &v2 = vertices[i2];

			// find normal
			OrientedPoint pt;
			pt.pos = (v0 + v1 + v2) / 3.0;
//			pt.norm = ~((v1 - v0) % (v2 - v0));
			pt.norm = ((v1 - v0) % (v2 - v0)) / 2;
			
			// sanity checks
#ifdef WIN32
			if (!_finite(pt.norm[0]) || !_finite(pt.norm[1]) || !_finite(pt.norm[2]))
#else
			if (!finite(pt.norm[0]) || !finite(pt.norm[1]) || !finite(pt.norm[2]))
#endif
			{
				printf("WARNING: degenerate triangle triangle for %d\n", tri_num);
				continue;
			}

			// add the point
			points.push_back(pt);
		}
	}

	file.close();

	findExtents(points, mine, maxe);
}

void PointSet::writePts(string filename)
{
	ofstream file(filename.c_str(), ios_base::binary);

	writeInline(file, mine);
	writeInline(file, maxe);
	writeInline(file, points);
	
	file.close();
}
